/*
 *  Copyright 2009 by Texas Instruments Incorporated.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>
 */

/*
 * irqk.c
 *
 * ============================================================================
 * Copyright (c) Texas Instruments Inc 2006
 *
 * Use of this software is controlled by the terms and conditions found in the
 * license agreement under which this software has been supplied or provided.
 * ============================================================================
 */
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/cdev.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/io.h>
#include <asm/memory.h>
#include <asm/uaccess.h>

#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#include <asm/arch/cpu.h>
#include <asm/arch/hardware.h>
#include <asm/arch/irqs.h>
#else
// the following are guesses for the upcoming LSP 2.10 for Linux 2.6.28
#include <mach/hardware.h>
#include <mach/cpu.h>
#include <mach/irqs.h>
#endif

#include "irqk.h"

/*
 * The following macros control version-dependent code:
 * USE_CACHE_VOID_ARG - #define if dmac functions take "void *" parameters,
 *    otherwise unsigned long is used
 * USE_CLASS_SIMPLE - #define if Linux version contains "class_simple*",
 *    otherwise "class*" or "device*" is used (see USE_CLASS_DEVICE usage).
 * USE_CLASS_DEVICE - #define if Linux version contains "class_device*",
 *    otherwise "device*" or "class_simple*" is used (see USE_CLASS_SIMPLE
 *    usage).
 * If neither USE_CLASS_SIMPLE nor USE_CLASS_DEVICE is set, there is further
 *    kernel version checking embedded in the module init & exit functions.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)

#warning *** not a warning *** Note: LINUX_VERSION_CODE >= 2.6.26

#undef USE_CLASS_DEVICE
#undef USE_CLASS_SIMPLE

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)

#warning *** not a warning *** Note: 2.6.26 > LINUX_VERSION_CODE >= 2.6.18

#define USE_CLASS_DEVICE
#undef USE_CLASS_SIMPLE

#else  /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18) */

#warning *** not a warning *** Note: LINUX_VERSION_CODE < 2.6.18

#define USE_CLASS_SIMPLE
#undef USE_CLASS_DEVICE

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18) */


/* A "channel" is an IRQ vector */
#define IRQK_NUMCHANNELS 64

/* holds physical address of ARM_INTMUX register */
unsigned int ARM_INTMUX;

/* one channel per IRQ vector on ARM */
struct channel {
    int transfer_complete;
    struct semaphore completion_sem;
    struct semaphore resource_sem;
    struct task_struct *owner;
};
static struct channel channels[IRQK_NUMCHANNELS];

struct intrinfo {
    int irq_num;
    int intrmux_bit;            // -1 if no bit in ARM_INTMUX register
    int intrmux_save;
    int intrmux_polarity;       // doesn't need to be set if intrmux_bit == -1
};
static struct intrinfo intr_info[IRQK_NUMRESOURCES][IRQK_NUMLINES];

/* tracks resource usage */
static struct intrinfo *registered_resource[IRQK_NUMRESOURCES] = {
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
};

#ifdef __DEBUG
#define __D(fmt, args...) printk(KERN_DEBUG "IRQK Debug: " fmt, ## args)
#else
#define __D(fmt, args...)
#endif

#define __E(fmt, args...) printk(KERN_ERR "IRQK Error: " fmt, ## args)

static int irq_major;
static atomic_t reference_count = ATOMIC_INIT(0);

#if (USE_UDEV==1)
#ifdef USE_CLASS_SIMPLE
static struct class_simple *irq_class;
#else
static struct class *irq_class;
#endif
#endif // USE_UDEV

static DECLARE_MUTEX(irq_mutex);

/* Forward declaration of system calls */
static int ioctl(struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long args);
static int open(struct inode *inode, struct file *filp);
static int release(struct inode *inode, struct file *filp);

static struct file_operations irq_fxns = {
    ioctl:   ioctl,
    open:    open,
    release: release
};

static irqreturn_t irqHandler(int irq, void *dev_id, struct pt_regs *regs)
{
    __D("irqHandler: servicing irq #%d\n", irq);

    up(&channels[irq].completion_sem);
    channels[irq].transfer_complete = 1;

    return IRQ_HANDLED;
}

/* releaseResource() must be called with irq_mutex held */
static void releaseResource(int resource)
{
    struct intrinfo *info;
    int mask = -1;
    int bit = 0;
    int intrmux;
    int irq;

    info = registered_resource[resource];
    irq = info->irq_num;

    channels[irq].transfer_complete = 0;
    channels[irq].owner = NULL;

    free_irq(irq, NULL);

    if (ARM_INTMUX) {
        if (info->intrmux_bit != -1) {
            __D("restoring ARM_INTMUX register ...\n");

            mask = 1 << info->intrmux_bit;
            bit = info->intrmux_save;
            intrmux = davinci_readl(ARM_INTMUX) & ~mask;
            intrmux |= bit;
            davinci_writel(intrmux, ARM_INTMUX);
        }
    }

    registered_resource[resource] = NULL;

    up(&channels[irq].resource_sem);
}


static int ioctl(struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long args)
{
    unsigned int __user *argp = (unsigned int __user *) args;
    IRQK_Request req;
    int resource;
    int irq;
    int rv;
    int mask = -1;
    int bit = 0;
    int intrmux;
    struct intrinfo *info;

    switch (cmd) {
        case IRQK_IOCREQUESTIRQ:
            if (copy_from_user(&req, argp, sizeof(IRQK_Request))) {
                return -EFAULT;
            }
            info = &intr_info[req.resource][req.flags & IRQK_INTRLINEMASK];
            irq = info->irq_num;

            if (irq == -1) {
                __D("REQUESTIRQ: no support for requested IRQ %d\n", req.resource);

                return -EINVAL;
            }

            __D("REQUESTIRQ ioctl received for irq #%d\n", irq);

            __D("acquiring irq's resource semaphore ...\n");

            if (down_interruptible(&channels[irq].resource_sem)) {
                return -ERESTARTSYS;
            }

            __D("... got it\n");

            down_interruptible(&irq_mutex);

            if ((rv = request_irq(irq, irqHandler, 0, "IRQK_Handler", NULL))) {
                __D("REQUESTIRQ: couldn't install ISR, "
                    "request_irq returned %d\n", rv);

                up(&irq_mutex);

                up(&channels[irq].resource_sem);

                return -EINVAL;
            }
            else {
                init_MUTEX_LOCKED(&channels[irq].completion_sem);
                channels[irq].transfer_complete = 0;
                channels[irq].owner = current;
                registered_resource[req.resource] = info;

                if (ARM_INTMUX) {
                    if (info->intrmux_bit != -1) {
                        __D("writing ARM_INTMUX register ...\n");

                        mask = 1 << info->intrmux_bit;
                        bit = info->intrmux_polarity << info->intrmux_bit;
                        intrmux = davinci_readl(ARM_INTMUX);
                        info->intrmux_save = intrmux & mask;
                        intrmux &= ~mask;
                        intrmux |= bit;
                        davinci_writel(intrmux, ARM_INTMUX);
                    }
                }
            }

            up(&irq_mutex);

            if (put_user(irq, argp)) {
                return -EFAULT;
            }

            break;

        case IRQK_IOCRELEASEIRQ:
            if (get_user(resource, argp)) {
                return -EFAULT;
            }

            info = registered_resource[resource];
            if (info == NULL) {
                __D("RELEASEIRQ: resource %d not registered\n", resource);

                return -EINVAL;
            }

            __D("RELEASEIRQ ioctl received for irq #%d\n", info->irq_num);

            down_interruptible(&irq_mutex);

            releaseResource(resource);

            up(&irq_mutex);

            break;

        case IRQK_IOCWAIT:
            if (get_user(resource, argp)) {
                return -EFAULT;
            }

            irq = registered_resource[resource]->irq_num;
            if (irq == -1) {
                __D("WAIT: resource %d not registered\n", resource);

                return -EINVAL;
            }

            __D("WAIT ioctl received for irq %d\n", irq);

            down_interruptible(&channels[irq].completion_sem);

            channels[irq].transfer_complete = 0;

            __D("...got it\n");

            break;

        case IRQK_IOCDONE:
            if (get_user(resource, argp)) {
                return -EFAULT;
            }

            irq = registered_resource[resource]->irq_num;
            if (irq == -1) {
                __D("DONE: resource %d not registered\n", resource);

                return -EINVAL;
            }

            __D("DONE ioctl received for irq #%d\n", irq);

            if (put_user(channels[irq].transfer_complete, argp)) {
                return -EFAULT;
            }

            break;

        default:
            __E("Unknown ioctl received: %d.\n", cmd);

            return -EINVAL;
    }

    return 0;
}

static int open(struct inode *inode, struct file *filp)
{
    __D("open: called.\n");

    atomic_inc(&reference_count);

    return 0;
}

static int release(struct inode *inode, struct file *filp)
{
    struct intrinfo *info;
    int irq;
    int i;

    __D("close: called.\n");

    down_interruptible(&irq_mutex);

    if (atomic_dec_and_test(&reference_count)) {
        __D("close: all references closed, force freeing busy IRQs ...\n");

        for (i = 0; i < IRQK_NUMRESOURCES; i++) {
            if (registered_resource[i] != NULL) {
                __D("close: ... freeing IRQ #%d\n", registered_resource[i]->irq_num);
                releaseResource(i);
            }
        }
    }
    else {
        /* unregister IRQs that are attached to the "current" process */
        for (i = 0; i < IRQK_NUMRESOURCES; i++) {
            if (registered_resource[i] != NULL) {
                info = registered_resource[i];
                irq = info->irq_num;
                if (channels[irq].owner == current) {
                    __D("close: force freeing busy IRQ #%d owned by current process.\n", irq);
                    releaseResource(i);
                }
            }
        }
    }

    up(&irq_mutex);

    return 0;
}

static void banner(void)
{
    printk(KERN_INFO "IRQK module: built on " __DATE__ " at " __TIME__ "\n");
    printk(KERN_INFO "  Reference Linux version %d.%d.%d\n",
           (LINUX_VERSION_CODE & 0x00ff0000) >> 16,
           (LINUX_VERSION_CODE & 0x0000ff00) >> 8,
           (LINUX_VERSION_CODE & 0x000000ff) >> 0
          );
    printk(KERN_INFO "  File " __FILE__ "\n");
}

static int __init irqk_init(void)
{
    int i;

    banner();

    ARM_INTMUX = 0;

    if (cpu_is_davinci_dm6467()) {
        __D("cpu is dm6467\n");
        printk(KERN_INFO "IRQK not supported for DM6467\n");

        return -EINVAL;
    }
    else if (cpu_is_davinci_dm355()) {
        __D("cpu is dm355\n");

        ARM_INTMUX = DM365_ARM_INTMUX;  // no macro for 355, but same as 360

        intr_info[IRQK_NSF][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_NSF][IRQK_FLEXIBLELINE].intrmux_bit = -1;
        // 11
        intr_info[IRQK_IMX0][IRQK_FLEXIBLELINE].irq_num = IRQ_DM3XX_IMCOPINT;
        intr_info[IRQK_IMX0][IRQK_FLEXIBLELINE].intrmux_bit = -1;
        intr_info[IRQK_IMX1][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_IMX1][IRQK_FLEXIBLELINE].intrmux_bit = -1;
        intr_info[IRQK_MJCP][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_MJCP][IRQK_FLEXIBLELINE].intrmux_bit = -1;
        intr_info[IRQK_HDVICP0][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_HDVICP0][IRQK_FLEXIBLELINE].intrmux_bit = -1;
        intr_info[IRQK_HDVICP1][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_HDVICP1][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        intr_info[IRQK_NSF][IRQK_FIXEDLINE].irq_num = -1;
        intr_info[IRQK_NSF][IRQK_FIXEDLINE].intrmux_bit = -1;
        // 10
        intr_info[IRQK_IMX0][IRQK_FIXEDLINE].irq_num = IRQ_IMXINT;
        intr_info[IRQK_IMX0][IRQK_FIXEDLINE].intrmux_bit = -1;
        intr_info[IRQK_IMX1][IRQK_FIXEDLINE].irq_num = -1;
        intr_info[IRQK_IMX1][IRQK_FIXEDLINE].intrmux_bit = -1;
        // 9
        intr_info[IRQK_MJCP][IRQK_FIXEDLINE].irq_num = IRQ_ASQINT;
        intr_info[IRQK_MJCP][IRQK_FIXEDLINE].intrmux_bit = -1;
        intr_info[IRQK_HDVICP0][IRQK_FIXEDLINE].irq_num = -1;
        intr_info[IRQK_HDVICP0][IRQK_FIXEDLINE].intrmux_bit = -1;
        intr_info[IRQK_HDVICP1][IRQK_FIXEDLINE].irq_num = -1;
        intr_info[IRQK_HDVICP1][IRQK_FIXEDLINE].intrmux_bit = -1;
    }
    else if (cpu_is_davinci_dm365()) {
        __D("cpu is dm365\n");

        ARM_INTMUX = DM365_ARM_INTMUX;

        intr_info[IRQK_NSF][IRQK_FLEXIBLELINE].irq_num = IRQ_DM3XX_IMCOPINT;
        intr_info[IRQK_NSF][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        intr_info[IRQK_IMX0][IRQK_FLEXIBLELINE].irq_num = IRQ_DM3XX_IMCOPINT;
        intr_info[IRQK_IMX0][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        intr_info[IRQK_IMX1][IRQK_FLEXIBLELINE].irq_num = IRQ_DM3XX_IMCOPINT;
        intr_info[IRQK_IMX1][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        // SEQ can't use "flex"
        intr_info[IRQK_MJCP][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_MJCP][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        intr_info[IRQK_HDVICP0][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_HDVICP0][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        intr_info[IRQK_HDVICP1][IRQK_FLEXIBLELINE].irq_num = -1;
        intr_info[IRQK_HDVICP1][IRQK_FLEXIBLELINE].intrmux_bit = -1;

        intr_info[IRQK_NSF][IRQK_FIXEDLINE].irq_num = IRQ_DM365_INSFINT;
        intr_info[IRQK_NSF][IRQK_FIXEDLINE].intrmux_bit = 25;
        intr_info[IRQK_NSF][IRQK_FIXEDLINE].intrmux_polarity = 1;

        intr_info[IRQK_IMX0][IRQK_FIXEDLINE].irq_num = IRQ_DM365_IMXINT0;
        intr_info[IRQK_IMX0][IRQK_FIXEDLINE].intrmux_bit = 0;
        intr_info[IRQK_IMX0][IRQK_FIXEDLINE].intrmux_polarity = 0;

        intr_info[IRQK_IMX1][IRQK_FIXEDLINE].irq_num = IRQ_DM365_IMXINT1;
        intr_info[IRQK_IMX1][IRQK_FIXEDLINE].intrmux_bit = 24;
        intr_info[IRQK_IMX1][IRQK_FIXEDLINE].intrmux_polarity = 1;

        intr_info[IRQK_MJCP][IRQK_FIXEDLINE].irq_num = IRQ_ASQINT;
        intr_info[IRQK_MJCP][IRQK_FIXEDLINE].intrmux_bit = -1;

        intr_info[IRQK_HDVICP0][IRQK_FIXEDLINE].irq_num = IRQ_DM365_KLD_ARMINT;
        intr_info[IRQK_HDVICP0][IRQK_FIXEDLINE].intrmux_bit = 0;
        intr_info[IRQK_HDVICP0][IRQK_FIXEDLINE].intrmux_polarity = 1;

        intr_info[IRQK_HDVICP1][IRQK_FIXEDLINE].irq_num = -1;
        intr_info[IRQK_HDVICP1][IRQK_FIXEDLINE].intrmux_bit = -1;
    }
    else if (cpu_is_davinci_dm644x()) {
        __D("cpu is dm644x\n");
        printk(KERN_INFO "IRQK not supported for DM644x\n");

        return -EINVAL;
    }
    else {
        __D("Unknown processor type\n");

        printk(KERN_INFO "Unknown processor type, IRQK not initialized\n");

        return -EINVAL;
    }

    irq_major = register_chrdev(0, "irqk", &irq_fxns);

    if (irq_major < 0) {
        __E("Failed to allocate major number.\n");
        return -ENODEV;
    }

    __D("Allocated major number: %d\n", irq_major);

#if (USE_UDEV==1)

#ifdef USE_CLASS_SIMPLE
    irq_class = class_simple_create(THIS_MODULE, "irqk");
#else
    irq_class = class_create(THIS_MODULE, "irqk");
#endif // USE_CLASS_SIMPLE

    if (IS_ERR(irq_class)) {
        __E("Error creating irq device class.\n");
        return -EIO;
    }

#ifdef USE_CLASS_SIMPLE
    class_simple_device_add(irq_class, MKDEV(irq_major, 0), NULL, "irqk");
#else
#ifdef USE_CLASS_DEVICE
    class_device_create(irq_class, NULL, MKDEV(irq_major, 0), NULL, "irqk");
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
    device_create(irq_class, NULL, MKDEV(irq_major, 0), NULL, "irqk");
#else
    device_create(irq_class, NULL, MKDEV(irq_major, 0), "irqk");
#endif // LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
#endif // USE_CLASS_DEVICE
#endif // USE_CLASS_SIMPLE

#endif // USE_UDEV==1

    for (i = 0; i < IRQK_NUMCHANNELS; i++) {
        init_MUTEX(&channels[i].resource_sem);
        channels[i].transfer_complete = 0;
        channels[i].owner = NULL;
    }

    printk(KERN_INFO "irqk initialized\n");

    __D("Successfully initialized module\n");

    return 0;
}

static void __exit irqk_exit(void)
{
    int i;

    __D("In irqk_exit()\n");

    for (i = 0; i < IRQK_NUMRESOURCES; i++) {
        if (registered_resource[i] != NULL) {
            __D("freeing registered irq %d\n", registered_resource[i]->irq_num);

            releaseResource(i);
        }
    }

#if (USE_UDEV==1)
#ifdef USE_CLASS_SIMPLE
    class_simple_device_remove(MKDEV(irq_major, 0));
    class_simple_destroy(irq_class);
#else
#ifdef USE_CLASS_DEVICE
    class_device_destroy(irq_class, MKDEV(irq_major, 0));
#else
    device_destroy(irq_class, MKDEV(irq_major, 0));
#endif
    class_destroy(irq_class);
#endif
#endif // USE_UDEV

    __D("Unregistering character device irqk\n");

    unregister_chrdev(irq_major, "irqk");

    printk("irqk unregistered\n");
}

module_init(irqk_init);
module_exit(irqk_exit);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("(c) Texas Instruments");

/*
 *  @(#) ti.sdo.linuxutils.irq; 1, 0, 0,44; 4-20-2009 11:43:52; /db/atree/library/trees/linuxutils/linuxutils-d06x/src/
 */

