/*
 * Copyright (c) 2009, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 *  @file       ti/sdo/linuxutils/vicp/src/interface/vicp.c
 *
 *  @brief      VICP Module interface Definitions (ARM) - 
 *              Allows algorithms to request and receive logical VICP resources
 *              The IRQK kernel module irqk.ko must be installed before
 *              using these APIS.  To install irqk.ko, enter
 *                  % insmod irqk.ko
 *              on a Linux command line.
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include <irqk.h>

#include "vicp.h"

#ifdef __DEBUG
#define __D(fmt, args...) fprintf(stderr, "VICP Debug: " fmt, ## args)
#else
#define __D(fmt, args...)
#endif

#define __E(fmt, args...) fprintf(stderr, "VICP Error: " fmt, ## args)

/* set VICP_irqk_fd to -2 to distinguish from failed open (-1) */
static int VICP_irqk_fd = -2;
static int VICP_refCount = 0;

static int validate_init()
{
    switch (VICP_irqk_fd) {
      case -3:
        __E("VICP_exit() already called, check stderr output for earlier "
            "VICP failure messages (possibly version mismatch).\n");

        return 0;

      case -2:
        __E("VICP_init() not called, you must initialize VICP before "
            "making API calls.\n");

        return 0;

      case -1:
        __E("IRQK file descriptor -1 (failed 'open()'), ensure IRQK "
            "kernel module irqk.ko has been installed with 'insmod'");

        return 0;

      default:
        return 1;
    }
}

int VICP_init() 
{
    int flags;

    if (VICP_refCount == 0) {
        VICP_irqk_fd = open("/dev/irqk", O_RDONLY);
        if (-1 == VICP_irqk_fd) {
            __E("init: Failed to open /dev/irqk: '%s'\n", strerror(errno));

            return -1;
        }
    }

    VICP_refCount++;

    __D("init: successfully opened /dev/irqk\n");

    flags = fcntl(VICP_irqk_fd, F_GETFD);
    if (flags != -1) {
        fcntl(VICP_irqk_fd, F_SETFD, flags | FD_CLOEXEC);
    }
    else {
        __E("init: fcntl(F_GETFD) failed: '%s'\n", strerror(errno));
    }

    return 0;
}

int VICP_exit() 
{
    VICP_refCount--;

    if (0 >= VICP_refCount) {
        if (-1 != VICP_irqk_fd) {
            close(VICP_irqk_fd);
        }
        /* setting -3 allows to distinguish IRQ exit from IRQ failed */
        VICP_irqk_fd = -3;
    }    

    return 0;
}

VICP_Status VICP_register(VICP_ResourceType resource,
                          VICP_InterruptLine intrLine,
                          VICP_InterruptType intrType)
{
    int rv;
    IRQK_Request req;

    if (!validate_init()) {
        return VICP_ENOINIT;
    }

    req.resource = resource;
    req.flags = intrLine == VICP_FIXED ? IRQK_FIXEDLINE : IRQK_FLEXIBLELINE;
    req.flags |= intrType == VICP_IRQ ? IRQK_IRQTYPE : IRQK_FIQTYPE;

    if ((rv = ioctl(VICP_irqk_fd, IRQK_IOCREQUESTIRQ, &req)) != 0) {
        __D("VICP_register() ioctl failed: %d\n", rv);

        return VICP_EFAIL;
    }

    return VICP_OK;
}

VICP_Status VICP_unregister(VICP_ResourceType resource)
{
    int rv;

    if (!validate_init()) {
        return VICP_ENOINIT;
    }

    if ((rv = ioctl(VICP_irqk_fd, IRQK_IOCRELEASEIRQ, &resource)) != 0) {
        __D("VICP_unregister() ioctl failed: %d\n", rv);

        return VICP_EFAIL;
    }

    return VICP_OK;
}

VICP_Status VICP_wait(VICP_ResourceType resource)
{
    int rv;

    if (!validate_init()) {
        return VICP_ENOINIT;
    }

    if ((rv = ioctl(VICP_irqk_fd, IRQK_IOCWAIT, &resource)) != 0) {
        __D("VICP_wait() ioctl failed: %d\n", rv);

        return VICP_EFAIL;
    }

    __D("VICP_wait() returned\n");

    return VICP_OK;
}

VICP_Status VICP_done(VICP_ResourceType resource, int *done)
{
    int rv;

    if (!validate_init()) {
        return VICP_ENOINIT;
    }

    if ((rv = ioctl(VICP_irqk_fd, IRQK_IOCDONE, &resource)) != 0) {
        __D("VICP_done() ioctl failed: %d\n", rv);

        return VICP_EFAIL;
    }

    *done = (int)resource;

    return VICP_OK;
}

/*
 *  @(#) ti.sdo.linuxutils.vicp; 1, 0, 0,44; 4-20-2009 11:44:15; /db/atree/library/trees/linuxutils/linuxutils-d06x/src/
 */

